@@ -1,18 +1,32 @@
-//! Safe interface for NumPy ndarray
-use crate::npyffi::{self, npy_intp, NPY_ORDER, PY_ARRAY_API};
-use ndarray::*;
+//! Safe interface for NumPy's ndarray class
+
+use std::{
+    marker::PhantomData,
+    mem,
+    os::raw::{c_int, c_void},
+    ptr, slice,
+};
+
+use ndarray::{
+    Array, ArrayBase, ArrayView, ArrayViewMut, Axis, Data, Dim, Dimension, IntoDimension, Ix0, Ix1,
+    Ix2, Ix3, Ix4, Ix5, Ix6, IxDyn, RawArrayView, RawArrayViewMut, RawData, Shape, ShapeBuilder,
+    StrideShape,
+};
 use num_traits::AsPrimitive;
 use pyo3::{
-    ffi, prelude::*, type_object, types::PyAny, AsPyPointer, PyDowncastError, PyNativeType,
-    PyResult,
+    ffi, pyobject_native_type_named, type_object, types::PyModule, AsPyPointer, FromPyObject,
+    IntoPy, Py, PyAny, PyDowncastError, PyErr, PyNativeType, PyObject, PyResult, PyTypeInfo,
+    Python, ToPyObject,
 };
-use std::{cell::Cell, mem, os::raw::c_int, ptr, slice};
-use std::{iter::ExactSizeIterator, marker::PhantomData};
 
-use crate::convert::{IntoPyArray, NpyIndex, ToNpyDims, ToPyArray};
-use crate::dtype::Element;
-use crate::error::{FromVecError, NotContiguousError, ShapeError};
-use crate::slice_box::SliceBox;
+use crate::convert::{ArrayExt, IntoPyArray, NpyIndex, ToNpyDims, ToPyArray};
+use crate::dtype::{Element, PyArrayDescr};
+use crate::error::{DimensionalityError, FromVecError, NotContiguousError, TypeError};
+use crate::npyffi::{self, npy_intp, NPY_ORDER, PY_ARRAY_API};
+#[allow(deprecated)]
+use crate::npyiter::{NpySingleIter, NpySingleIterBuilder, ReadWrite};
+use crate::readonly::PyReadonlyArray;
+use crate::slice_container::PySliceContainer;
 
 /// A safe, static-typed interface for
 /// [NumPy ndarray](https://numpy.org/doc/stable/reference/arrays.ndarray.html).
@@ -94,22 +108,30 @@ pub type PyArray6<T> = PyArray<T, Ix6>;
 /// Dynamic-dimensional array.
 pub type PyArrayDyn<T> = PyArray<T, IxDyn>;
 
-/// Returns a array module.
+/// Returns a handle to NumPy's multiarray module.
 pub fn get_array_module(py: Python<'_>) -> PyResult<&PyModule> {
     PyModule::import(py, npyffi::array::MOD_NAME)
 }
 
 unsafe impl<T, D> type_object::PyLayout<PyArray<T, D>> for npyffi::PyArrayObject {}
+
 impl<T, D> type_object::PySizedLayout<PyArray<T, D>> for npyffi::PyArrayObject {}
 
-pyobject_native_type_info!(
-    PyArray<T, D>,
-    *npyffi::PY_ARRAY_API.get_type_object(npyffi::NpyTypes::PyArray_Type),
-    Some("numpy"),
-    #checkfunction=npyffi::PyArray_Check
-    ; T
-    ; D
-);
+unsafe impl<T: Element, D: Dimension> PyTypeInfo for PyArray<T, D> {
+    type AsRefTarget = Self;
+
+    const NAME: &'static str = "PyArray<T, D>";
+    const MODULE: ::std::option::Option<&'static str> = Some("numpy");
+
+    #[inline]
+    fn type_object_raw(py: Python) -> *mut ffi::PyTypeObject {
+        unsafe { npyffi::PY_ARRAY_API.get_type_object(py, npyffi::NpyTypes::PyArray_Type) }
+    }
+
+    fn is_type_of(ob: &PyAny) -> bool {
+        <&Self>::extract(ob).is_ok()
+    }
+}
 
 pyobject_native_type_named!(PyArray<T, D> ; T ; D);
 
@@ -119,25 +141,33 @@ impl<T, D> IntoPy<PyObject> for PyArray<T, D> {
     }
 }
 
-impl<'a, T: Element, D: Dimension> FromPyObject<'a> for &'a PyArray<T, D> {
+impl<'py, T: Element, D: Dimension> FromPyObject<'py> for &'py PyArray<T, D> {
     // here we do type-check three times
     // 1. Checks if the object is PyArray
     // 2. Checks if the data type of the array is T
     // 3. Checks if the dimension is same as D
-    fn extract(ob: &'a PyAny) -> PyResult<Self> {
+    fn extract(ob: &'py PyAny) -> PyResult<Self> {
         let array = unsafe {
-            if npyffi::PyArray_Check(ob.as_ptr()) == 0 {
+            if npyffi::PyArray_Check(ob.py(), ob.as_ptr()) == 0 {
                 return Err(PyDowncastError::new(ob, "PyArray<T, D>").into());
             }
             &*(ob as *const PyAny as *const PyArray<T, D>)
         };
-        let dtype = array.dtype();
-        let dim = array.shape().len();
-        if T::is_same_type(dtype) && D::NDIM.map(|n| n == dim).unwrap_or(true) {
-            Ok(array)
-        } else {
-            Err(ShapeError::new(dtype, dim, T::DATA_TYPE, D::NDIM).into())
+
+        let src_dtype = array.dtype();
+        let dst_dtype = T::get_dtype(ob.py());
+        if !src_dtype.is_equiv_to(dst_dtype) {
+            return Err(TypeError::new(src_dtype, dst_dtype).into());
         }
+
+        let src_ndim = array.shape().len();
+        if let Some(dst_ndim) = D::NDIM {
+            if src_ndim != dst_ndim {
+                return Err(DimensionalityError::new(src_ndim, dst_ndim).into());
+            }
+        }
+
+        Ok(array)
     }
 }
 
@@ -155,10 +185,10 @@ impl<T, D> PyArray<T, D> {
     /// pyo3::Python::with_gil(|py| {
     ///    let array = numpy::PyArray::from_vec(py, vec![1, 2, 3i32]);
     ///    let dtype = array.dtype();
-    ///    assert_eq!(dtype.get_datatype().unwrap(), numpy::DataType::Int32);
+    ///    assert!(dtype.is_equiv_to(numpy::dtype::<i32>(py)));
     /// });
     /// ```
-    pub fn dtype(&self) -> &crate::PyArrayDescr {
+    pub fn dtype(&self) -> &PyArrayDescr {
         let descr_ptr = unsafe { (*self.as_array_ptr()).descr };
         unsafe { pyo3::FromPyPointer::from_borrowed_ptr(self.py(), descr_ptr as _) }
     }
@@ -174,7 +204,7 @@ impl<T, D> PyArray<T, D> {
     }
 
     /// Returns a temporally unwriteable reference of the array.
-    pub fn readonly(&self) -> crate::PyReadonlyArray<T, D> {
+    pub fn readonly(&self) -> PyReadonlyArray<T, D> {
         self.into()
     }
 
@@ -189,7 +219,7 @@ impl<T, D> PyArray<T, D> {
     ///     assert!(array.is_contiguous());
     ///     let locals = [("np", numpy::get_array_module(py).unwrap())].into_py_dict(py);
     ///     let not_contiguous: &numpy::PyArray1<f32> = py
-    ///         .eval("np.zeros((3, 5))[::2, 4]", Some(locals), None)
+    ///         .eval("np.zeros((3, 5), dtype='float32')[::2, 4]", Some(locals), None)
     ///         .unwrap()
     ///         .downcast()
     ///         .unwrap();
@@ -231,13 +261,24 @@ impl<T, D> PyArray<T, D> {
         unsafe { Py::from_borrowed_ptr(self.py(), self.as_ptr()) }
     }
 
-    /// Constructs `PyArray` from raw python object without incrementing reference counts.
+    /// Constructs `PyArray` from raw Python object without incrementing reference counts.
+    ///
+    /// # Safety
+    ///
+    /// Implementations must ensure the object does not get freed during `'py`
+    /// and ensure that `ptr` is of the correct type.
     pub unsafe fn from_owned_ptr(py: Python<'_>, ptr: *mut ffi::PyObject) -> &Self {
         py.from_owned_ptr(ptr)
     }
 
-    /// Constructs PyArray from raw python object and increments reference counts.
-    pub unsafe fn from_borrowed_ptr(py: Python<'_>, ptr: *mut ffi::PyObject) -> &Self {
+    /// Constructs PyArray from raw Python object and increments reference counts.
+    ///
+    /// # Safety
+    ///
+    /// Implementations must ensure the object does not get freed during `'py`
+    /// and ensure that `ptr` is of the correct type.
+    /// Note that it must be safe to decrement the reference count of ptr.
+    pub unsafe fn from_borrowed_ptr<'py>(py: Python<'py>, ptr: *mut ffi::PyObject) -> &'py Self {
         py.from_borrowed_ptr(ptr)
     }
 
@@ -249,7 +290,7 @@ impl<T, D> PyArray<T, D> {
     /// ```
     /// use numpy::PyArray3;
     /// pyo3::Python::with_gil(|py| {
-    ///     let arr = PyArray3::<f64>::new(py, [4, 5, 6], false);
+    ///     let arr = PyArray3::<f64>::zeros(py, [4, 5, 6], false);
     ///     assert_eq!(arr.ndim(), 3);
     /// });
     /// ```
@@ -266,7 +307,7 @@ impl<T, D> PyArray<T, D> {
     /// ```
     /// use numpy::PyArray3;
     /// pyo3::Python::with_gil(|py| {
-    ///     let arr = PyArray3::<f64>::new(py, [4, 5, 6], false);
+    ///     let arr = PyArray3::<f64>::zeros(py, [4, 5, 6], false);
     ///     assert_eq!(arr.strides(), &[240, 48, 8]);
     /// });
     /// ```
@@ -287,7 +328,7 @@ impl<T, D> PyArray<T, D> {
     /// ```
     /// use numpy::PyArray3;
     /// pyo3::Python::with_gil(|py| {
-    ///     let arr = PyArray3::<f64>::new(py, [4, 5, 6], false);
+    ///     let arr = PyArray3::<f64>::zeros(py, [4, 5, 6], false);
     ///     assert_eq!(arr.shape(), &[4, 5, 6]);
     /// });
     /// ```
@@ -321,12 +362,25 @@ impl<T, D> PyArray<T, D> {
     }
 }
 
-struct InvertedAxises(Vec<Axis>);
+struct InvertedAxes(u32);
 
-impl InvertedAxises {
-    fn invert<S: RawData, D: Dimension>(self, array: &mut ArrayBase<S, D>) {
-        for axis in self.0 {
-            array.invert_axis(axis);
+impl InvertedAxes {
+    fn new(len: usize) -> Self {
+        assert!(len <= 32, "Only dimensionalities of up to 32 are supported");
+        Self(0)
+    }
+
+    fn push(&mut self, axis: usize) {
+        debug_assert!(axis < 32);
+        self.0 |= 1 << axis;
+    }
+
+    fn invert<S: RawData, D: Dimension>(mut self, array: &mut ArrayBase<S, D>) {
+        while self.0 != 0 {
+            let axis = self.0.trailing_zeros() as usize;
+            self.0 &= !(1 << axis);
+
+            array.invert_axis(Axis(axis));
         }
     }
 }
@@ -335,56 +389,85 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
     /// Same as [shape](#method.shape), but returns `D`
     #[inline(always)]
     pub fn dims(&self) -> D {
-        D::from_dimension(&Dim(self.shape())).expect("PyArray::dims different dimension")
+        D::from_dimension(&Dim(self.shape())).expect("mismatching dimensions")
     }
 
-    fn ndarray_shape_ptr(&self) -> (StrideShape<D>, *mut T, InvertedAxises) {
-        const ERR_MSG: &str = "PyArray::ndarray_shape: dimension mismatching";
-        let shape_slice = self.shape();
-        let shape: Shape<_> = Dim(self.dims()).into();
-        let sizeof_t = mem::size_of::<T>();
+    fn ndarray_shape_ptr(&self) -> (StrideShape<D>, *mut T, InvertedAxes) {
+        let shape = self.shape();
         let strides = self.strides();
+
         let mut new_strides = D::zeros(strides.len());
         let mut data_ptr = unsafe { self.data() };
-        let mut inverted_axises = vec![];
+        let mut inverted_axes = InvertedAxes::new(strides.len());
+
         for i in 0..strides.len() {
             // TODO(kngwyu): Replace this hacky negative strides support with
             // a proper constructor, when it's implemented.
             // See https://github.com/rust-ndarray/ndarray/issues/842 for more.
             if strides[i] < 0 {
                 // Move the pointer to the start position
-                let offset = strides[i] * (shape_slice[i] as isize - 1) / sizeof_t as isize;
+                let offset = strides[i] * (shape[i] as isize - 1) / mem::size_of::<T>() as isize;
                 unsafe {
                     data_ptr = data_ptr.offset(offset);
                 }
-                new_strides[i] = (-strides[i]) as usize / sizeof_t;
-                inverted_axises.push(Axis(i));
+                new_strides[i] = (-strides[i]) as usize / mem::size_of::<T>();
+
+                inverted_axes.push(i);
             } else {
-                new_strides[i] = strides[i] as usize / sizeof_t;
+                new_strides[i] = strides[i] as usize / mem::size_of::<T>();
             }
         }
-        let st = D::from_dimension(&Dim(new_strides)).expect(ERR_MSG);
-        (shape.strides(st), data_ptr, InvertedAxises(inverted_axises))
+
+        let shape = Shape::from(D::from_dimension(&Dim(shape)).expect("mismatching dimensions"));
+        let new_strides = D::from_dimension(&Dim(new_strides)).expect("mismatching dimensions");
+
+        (shape.strides(new_strides), data_ptr, inverted_axes)
     }
 
     /// Creates a new uninitialized PyArray in python heap.
     ///
     /// If `is_fortran == true`, returns Fortran-order array. Else, returns C-order array.
     ///
+    /// # Safety
+    ///
+    /// The returned array will always be safe to be dropped as the elements must either
+    /// be trivially copyable or have `DATA_TYPE == DataType::Object`, i.e. be pointers
+    /// into Python's heap, which NumPy will automatically zero-initialize.
+    ///
+    /// However, the elements themselves will not be valid and should only be accessed
+    /// via raw pointers obtained via [uget_raw](#method.uget_raw).
+    ///
+    /// All methods which produce references to the elements invoke undefined behaviour.
+    /// In particular, zero-initialized pointers are _not_ valid instances of `PyObject`.
+    ///
     /// # Example
     /// ```
     /// use numpy::PyArray3;
+    ///
     /// pyo3::Python::with_gil(|py| {
-    ///     let arr = PyArray3::<i32>::new(py, [4, 5, 6], false);
+    ///     let arr = unsafe {
+    ///         let arr = PyArray3::<i32>::new(py, [4, 5, 6], false);
+    ///
+    ///         for i in 0..4 {
+    ///             for j in 0..5 {
+    ///                 for k in 0..6 {
+    ///                     arr.uget_raw([i, j, k]).write((i * j * k) as i32);
+    ///                 }
+    ///             }
+    ///         }
+    ///
+    ///         arr
+    ///     };
+    ///
     ///     assert_eq!(arr.shape(), &[4, 5, 6]);
     /// });
     /// ```
-    pub fn new<ID>(py: Python, dims: ID, is_fortran: bool) -> &Self
+    pub unsafe fn new<ID>(py: Python, dims: ID, is_fortran: bool) -> &Self
     where
         ID: IntoDimension<Dim = D>,
     {
         let flags = if is_fortran { 1 } else { 0 };
-        unsafe { PyArray::new_(py, dims, ptr::null_mut(), flags) }
+        PyArray::new_(py, dims, ptr::null_mut(), flags)
     }
 
     pub(crate) unsafe fn new_<ID>(
@@ -397,56 +480,131 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
         ID: IntoDimension<Dim = D>,
     {
         let dims = dims.into_dimension();
-        let ptr = PY_ARRAY_API.PyArray_New(
-            PY_ARRAY_API.get_type_object(npyffi::NpyTypes::PyArray_Type),
+        let ptr = PY_ARRAY_API.PyArray_NewFromDescr(
+            py,
+            PY_ARRAY_API.get_type_object(py, npyffi::NpyTypes::PyArray_Type),
+            T::get_dtype(py).into_dtype_ptr(),
             dims.ndim_cint(),
             dims.as_dims_ptr(),
-            T::npy_type() as i32,
-            strides as *mut _, // strides
-            ptr::null_mut(),   // data
-            0,                 // itemsize
-            flag,              // flag
-            ptr::null_mut(),   // obj
+            strides as *mut npy_intp, // strides
+            ptr::null_mut(),          // data
+            flag,                     // flag
+            ptr::null_mut(),          // obj
         );
         Self::from_owned_ptr(py, ptr)
     }
 
-    pub(crate) unsafe fn from_boxed_slice<'py, ID>(
+    unsafe fn new_with_data<'py, ID>(
         py: Python<'py>,
         dims: ID,
         strides: *const npy_intp,
-        boxed_slice: Box<[T]>,
-        data_ptr: Option<*const T>,
+        data_ptr: *const T,
+        container: *mut PyAny,
     ) -> &'py Self
     where
         ID: IntoDimension<Dim = D>,
     {
         let dims = dims.into_dimension();
-        let container = SliceBox::new(boxed_slice);
-        let data_ptr = data_ptr.unwrap_or_else(|| container.data.as_ptr());
-        let cell = pyo3::PyClassInitializer::from(container)
-            .create_cell(py)
-            .expect("Object creation failed.");
-        let ptr = PY_ARRAY_API.PyArray_New(
-            PY_ARRAY_API.get_type_object(npyffi::NpyTypes::PyArray_Type),
+        let ptr = PY_ARRAY_API.PyArray_NewFromDescr(
+            py,
+            PY_ARRAY_API.get_type_object(py, npyffi::NpyTypes::PyArray_Type),
+            T::get_dtype(py).into_dtype_ptr(),
             dims.ndim_cint(),
             dims.as_dims_ptr(),
-            T::npy_type() as i32,
-            strides as *mut _,           // strides
-            data_ptr as _,               // data
-            mem::size_of::<T>() as i32,  // itemsize
+            strides as *mut npy_intp,    // strides
+            data_ptr as *mut c_void,     // data
             npyffi::NPY_ARRAY_WRITEABLE, // flag
             ptr::null_mut(),             // obj
         );
-        PY_ARRAY_API.PyArray_SetBaseObject(ptr as *mut npyffi::PyArrayObject, cell as _);
+
+        PY_ARRAY_API.PyArray_SetBaseObject(
+            py,
+            ptr as *mut npyffi::PyArrayObject,
+            container as *mut ffi::PyObject,
+        );
+
         Self::from_owned_ptr(py, ptr)
     }
 
+    pub(crate) unsafe fn from_raw_parts<'py, ID, C>(
+        py: Python<'py>,
+        dims: ID,
+        strides: *const npy_intp,
+        data_ptr: *const T,
+        container: C,
+    ) -> &'py Self
+    where
+        ID: IntoDimension<Dim = D>,
+        PySliceContainer: From<C>,
+    {
+        let container = pyo3::PyClassInitializer::from(PySliceContainer::from(container))
+            .create_cell(py)
+            .expect("Object creation failed.");
+
+        Self::new_with_data(py, dims, strides, data_ptr, container as *mut PyAny)
+    }
+
+    /// Creates a NumPy array backed by `array` and ties its ownership to the Python object `container`.
+    ///
+    /// # Safety
+    ///
+    /// `container` is set as a base object of the returned array which must not be dropped until `container` is dropped.
+    /// Furthermore, `array` must not be reallocated from the time this method is called and until `container` is dropped.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// # use pyo3::prelude::*;
+    /// # use numpy::{ndarray::Array1, PyArray1};
+    /// #
+    /// #[pyclass]
+    /// struct Owner {
+    ///     array: Array1<f64>,
+    /// }
+    ///
+    /// #[pymethods]
+    /// impl Owner {
+    ///     #[getter]
+    ///     fn array<'py>(this: &'py PyCell<Self>) -> &'py PyArray1<f64> {
+    ///         let array = &this.borrow().array;
+    ///
+    ///         // SAFETY: The memory backing `array` will stay valid as long as this object is alive
+    ///         // as we do not modify `array` in any way which would cause it to be reallocated.
+    ///         unsafe { PyArray1::borrow_from_array(array, this) }
+    ///     }
+    /// }
+    /// ```
+    pub unsafe fn borrow_from_array<'py, S>(
+        array: &ArrayBase<S, D>,
+        container: &'py PyAny,
+    ) -> &'py Self
+    where
+        S: Data<Elem = T>,
+    {
+        let (strides, dims) = (array.npy_strides(), array.raw_dim());
+        let data_ptr = array.as_ptr();
+
+        let py = container.py();
+
+        mem::forget(container.to_object(py));
+
+        Self::new_with_data(
+            py,
+            dims,
+            strides.as_ptr(),
+            data_ptr,
+            container as *const PyAny as *mut PyAny,
+        )
+    }
+
     /// Construct a new nd-dimensional array filled with 0.
     ///
     /// If `is_fortran` is true, then
     /// a fortran order array is created, otherwise a C-order array is created.
     ///
+    /// For elements with `DATA_TYPE == DataType::Object`, this will fill the array
+    /// with valid pointers to zero-valued Python integer objects.
+    ///
     /// See also [PyArray_Zeros](https://numpy.org/doc/stable/reference/c-api/array.html#c.PyArray_Zeros)
     ///
     /// # Example
@@ -464,11 +622,11 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
     {
         let dims = dims.into_dimension();
         unsafe {
-            let dtype = T::get_dtype(py);
             let ptr = PY_ARRAY_API.PyArray_Zeros(
+                py,
                 dims.ndim_cint(),
                 dims.as_dims_ptr(),
-                dtype.into_ptr() as _,
+                T::get_dtype(py).into_dtype_ptr(),
                 if is_fortran { -1 } else { 0 },
             );
             Self::from_owned_ptr(py, ptr)
@@ -477,9 +635,7 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
 
     /// Returns the immutable view of the internal data of `PyArray` as slice.
     ///
-    /// Please consider the use of safe alternatives
-    /// ([`PyReadonlyArray::as_slice`](../struct.PyReadonlyArray.html#method.as_slice)
-    /// , [`as_cell_slice`](#method.as_cell_slice) or [`to_vec`](#method.to_vec)) instead of this.
+    /// Please consider the use of the safe alternative [`PyReadonlyArray::as_slice`].
     ///
     /// # Safety
     /// If the internal array is not readonly and can be mutated from Python code,
@@ -492,22 +648,11 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
         }
     }
 
-    /// Returns the view of the internal data of `PyArray` as `&[Cell<T>]`.
-    pub fn as_cell_slice(&self) -> Result<&[Cell<T>], NotContiguousError> {
-        if !self.is_contiguous() {
-            Err(NotContiguousError)
-        } else {
-            Ok(unsafe { slice::from_raw_parts(self.data() as _, self.len()) })
-        }
-    }
-
     /// Returns the view of the internal data of `PyArray` as mutable slice.
     ///
     /// # Safety
     /// If another reference to the internal data exists(e.g., `&[T]` or `ArrayView`),
     /// it might cause undefined behavior.
-    ///
-    /// In such case, please consider the use of [`as_cell_slice`](#method.as_cell_slice),
     pub unsafe fn as_slice_mut(&self) -> Result<&mut [T], NotContiguousError> {
         if !self.is_contiguous() {
             Err(NotContiguousError)
@@ -516,23 +661,25 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
         }
     }
 
-    /// Construct PyArray from
-    /// [`ndarray::Array`](https://docs.rs/ndarray/latest/ndarray/type.Array.html).
+    /// Constructs a `PyArray` from [`ndarray::Array`]
     ///
-    /// This method uses internal [`Vec`](https://doc.rust-lang.org/std/vec/struct.Vec.html)
-    /// of `ndarray::Array` as numpy array.
+    /// This method uses the internal [`Vec`] of the `ndarray::Array` as the base object of the NumPy array.
     ///
     /// # Example
+    ///
     /// ```
-    /// # #[macro_use] extern crate ndarray;
+    /// use ndarray::array;
     /// use numpy::PyArray;
+    ///
     /// pyo3::Python::with_gil(|py| {
     ///     let pyarray = PyArray::from_owned_array(py, array![[1, 2], [3, 4]]);
     ///     assert_eq!(pyarray.readonly().as_array(), array![[1, 2], [3, 4]]);
     /// });
     /// ```
     pub fn from_owned_array<'py>(py: Python<'py>, arr: Array<T, D>) -> &'py Self {
-        IntoPyArray::into_pyarray(arr, py)
+        let (strides, dims) = (arr.npy_strides(), arr.raw_dim());
+        let data_ptr = arr.as_ptr();
+        unsafe { PyArray::from_raw_parts(py, dims, strides.as_ptr(), data_ptr, arr) }
     }
 
     /// Get the immutable reference of the specified element, with checking the passed index is valid.
@@ -563,7 +710,10 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
     ///
     /// See [NpyIndex](../convert/trait.NpyIndex.html) for what types you can use as index.
     ///
-    /// Passing an invalid index can cause undefined behavior(mostly SIGSEGV).
+    /// # Safety
+    ///
+    /// Passing an invalid index is undefined behavior. The element must also have been initialized.
+    /// The elemet must also not be modified by Python code.
     ///
     /// # Example
     /// ```
@@ -583,6 +733,11 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
     }
 
     /// Same as [uget](#method.uget), but returns `&mut T`.
+    ///
+    /// # Safety
+    ///
+    /// Passing an invalid index is undefined behavior. The element must also have been initialized.
+    /// The element must also not be accessed by Python code.
     #[inline(always)]
     #[allow(clippy::mut_from_ref)]
     pub unsafe fn uget_mut<Idx>(&self, index: Idx) -> &mut T
@@ -593,6 +748,19 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
         &mut *(self.data().offset(offset) as *mut _)
     }
 
+    /// Same as [uget](#method.uget), but returns `*mut T`.
+    ///
+    /// # Safety
+    /// Passing an invalid index is undefined behavior.
+    #[inline(always)]
+    pub unsafe fn uget_raw<Idx>(&self, index: Idx) -> *mut T
+    where
+        Idx: NpyIndex<Dim = D>,
+    {
+        let offset = index.get_unchecked::<T>(self.strides());
+        self.data().offset(offset) as *mut _
+    }
+
     /// Get dynamic dimensioned array from fixed dimension array.
     pub fn to_dyn(&self) -> &PyArray<T, IxDyn> {
         let python = self.py();
@@ -670,21 +838,37 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
     /// If the internal array is not readonly and can be mutated from Python code,
     /// holding the `ArrayView` might cause undefined behavior.
     pub unsafe fn as_array(&self) -> ArrayView<'_, T, D> {
-        let (shape, ptr, inverted_axises) = self.ndarray_shape_ptr();
+        let (shape, ptr, inverted_axes) = self.ndarray_shape_ptr();
         let mut res = ArrayView::from_shape_ptr(shape, ptr);
-        inverted_axises.invert(&mut res);
+        inverted_axes.invert(&mut res);
         res
     }
 
-    /// Returns the internal array as `ArrayViewMut`. See also [`as_array`](#method.as_array).
+    /// Returns the internal array as [`ArrayViewMut`]. See also [`as_array`](#method.as_array).
     ///
     /// # Safety
     /// If another reference to the internal data exists(e.g., `&[T]` or `ArrayView`),
     /// it might cause undefined behavior.
     pub unsafe fn as_array_mut(&self) -> ArrayViewMut<'_, T, D> {
-        let (shape, ptr, inverted_axises) = self.ndarray_shape_ptr();
+        let (shape, ptr, inverted_axes) = self.ndarray_shape_ptr();
         let mut res = ArrayViewMut::from_shape_ptr(shape, ptr);
-        inverted_axises.invert(&mut res);
+        inverted_axes.invert(&mut res);
+        res
+    }
+
+    /// Returns the internal array as [`RawArrayView`] enabling element access via raw pointers
+    pub fn as_raw_array(&self) -> RawArrayView<T, D> {
+        let (shape, ptr, inverted_axes) = self.ndarray_shape_ptr();
+        let mut res = unsafe { RawArrayView::from_shape_ptr(shape, ptr) };
+        inverted_axes.invert(&mut res);
+        res
+    }
+
+    /// Returns the internal array as [`RawArrayViewMut`] enabling element access via raw pointers
+    pub fn as_raw_array_mut(&self) -> RawArrayViewMut<T, D> {
+        let (shape, ptr, inverted_axes) = self.ndarray_shape_ptr();
+        let mut res = unsafe { RawArrayViewMut::from_shape_ptr(shape, ptr) };
+        inverted_axes.invert(&mut res);
         res
     }
 
@@ -708,6 +892,48 @@ impl<T: Element, D: Dimension> PyArray<T, D> {
     }
 }
 
+impl<D: Dimension> PyArray<PyObject, D> {
+    /// Constructs a `PyArray` containing objects from [`ndarray::Array`]
+    ///
+    /// This method uses the internal [`Vec`] of the `ndarray::Array` as the base object the NumPy array.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use ndarray::array;
+    /// use pyo3::{pyclass, Py, Python};
+    /// use numpy::PyArray;
+    ///
+    /// #[pyclass]
+    /// struct CustomElement {
+    ///     foo: i32,
+    ///     bar: f64,
+    /// }
+    ///
+    /// Python::with_gil(|py| {
+    ///     let array = array![
+    ///         Py::new(py, CustomElement {
+    ///             foo: 1,
+    ///             bar: 2.0,
+    ///         }).unwrap(),
+    ///         Py::new(py, CustomElement {
+    ///             foo: 3,
+    ///             bar: 4.0,
+    ///         }).unwrap(),
+    ///     ];
+    ///
+    ///     let pyarray = PyArray::from_owned_object_array(py, array);
+    ///
+    ///     assert!(pyarray.readonly().get(0).unwrap().as_ref(py).is_instance_of::<CustomElement>().unwrap());
+    /// });
+    /// ```
+    pub fn from_owned_object_array<'py, T>(py: Python<'py>, arr: Array<Py<T>, D>) -> &'py Self {
+        let (strides, dims) = (arr.npy_strides(), arr.raw_dim());
+        let data_ptr = arr.as_ptr() as *const PyObject;
+        unsafe { PyArray::from_raw_parts(py, dims, strides.as_ptr(), data_ptr, arr) }
+    }
+}
+
 impl<T: Copy + Element> PyArray<T, Ix0> {
     /// Get the element of zero-dimensional PyArray.
     ///
@@ -730,11 +956,18 @@ impl<T: Element> PyArray<T, Ix1> {
     /// });
     /// ```
     pub fn from_slice<'py>(py: Python<'py>, slice: &[T]) -> &'py Self {
-        let array = PyArray::new(py, [slice.len()], false);
         unsafe {
-            array.copy_ptr(slice.as_ptr(), slice.len());
+            let array = PyArray::new(py, [slice.len()], false);
+            if T::IS_COPY {
+                array.copy_ptr(slice.as_ptr(), slice.len());
+            } else {
+                let data_ptr = array.data();
+                for (i, item) in slice.iter().enumerate() {
+                    data_ptr.add(i).write(item.clone());
+                }
+            }
+            array
         }
-        array
     }
 
     /// Construct one-dimension PyArray
@@ -767,13 +1000,20 @@ impl<T: Element> PyArray<T, Ix1> {
     /// });
     /// ```
     pub fn from_exact_iter(py: Python<'_>, iter: impl ExactSizeIterator<Item = T>) -> &Self {
-        let array = Self::new(py, [iter.len()], false);
+        // NumPy will always zero-initialize object pointers,
+        // so the array can be dropped safely if the iterator panics.
         unsafe {
-            for (i, item) in iter.enumerate() {
-                *array.uget_mut([i]) = item;
+            let len = iter.len();
+            let array = Self::new(py, [len], false);
+            let mut idx = 0;
+            for item in iter {
+                assert!(idx < len);
+                array.uget_raw([idx]).write(item);
+                idx += 1;
             }
+            assert!(idx == len);
+            array
         }
-        array
     }
 
     /// Construct one-dimension PyArray from a type which implements
@@ -795,9 +1035,11 @@ impl<T: Element> PyArray<T, Ix1> {
         let iter = iter.into_iter();
         let (min_len, max_len) = iter.size_hint();
         let mut capacity = max_len.unwrap_or_else(|| min_len.max(512 / mem::size_of::<T>()));
-        let array = Self::new(py, [capacity], false);
-        let mut length = 0;
         unsafe {
+            // NumPy will always zero-initialize object pointers,
+            // so the array can be dropped safely if the iterator panics.
+            let array = Self::new(py, [capacity], false);
+            let mut length = 0;
             for (i, item) in iter.enumerate() {
                 length += 1;
                 if length > capacity {
@@ -806,13 +1048,13 @@ impl<T: Element> PyArray<T, Ix1> {
                         .resize(capacity)
                         .expect("PyArray::from_iter: Failed to allocate memory");
                 }
-                *array.uget_mut([i]) = item;
+                array.uget_raw([i]).write(item);
             }
+            if capacity > length {
+                array.resize(length).unwrap()
+            }
+            array
         }
-        if capacity > length {
-            array.resize(length).unwrap()
-        }
-        array
     }
 
     /// Extends or trancates the length of 1 dimension PyArray.
@@ -828,19 +1070,27 @@ impl<T: Element> PyArray<T, Ix1> {
     /// });
     /// ```
     pub fn resize(&self, new_elems: usize) -> PyResult<()> {
-        self.resize_([new_elems], 1, NPY_ORDER::NPY_ANYORDER)
+        self.resize_(self.py(), [new_elems], 1, NPY_ORDER::NPY_ANYORDER)
     }
 
     /// Iterates all elements of this array.
     /// See [NpySingleIter](../npyiter/struct.NpySingleIter.html) for more.
-    pub fn iter<'py>(
-        &'py self,
-    ) -> PyResult<crate::NpySingleIter<'py, T, crate::npyiter::ReadWrite>> {
-        crate::NpySingleIterBuilder::readwrite(self).build()
+    ///
+    /// # Safety
+    ///
+    /// The iterator will produce mutable references into the array which must not be
+    /// aliased by other references for the life time of the iterator.
+    #[deprecated(
+        note = "The wrappers of the array iterator API are deprecated, please use ndarray's `ArrayBase::iter_mut` instead."
+    )]
+    #[allow(deprecated)]
+    pub unsafe fn iter<'py>(&'py self) -> PyResult<NpySingleIter<'py, T, ReadWrite>> {
+        NpySingleIterBuilder::readwrite(self).build()
     }
 
     fn resize_<D: IntoDimension>(
         &self,
+        py: Python,
         dims: D,
         check_ref: c_int,
         order: NPY_ORDER,
@@ -849,6 +1099,7 @@ impl<T: Element> PyArray<T, Ix1> {
         let mut np_dims = dims.to_npy_dims();
         let res = unsafe {
             PY_ARRAY_API.PyArray_Resize(
+                py,
                 self.as_array_ptr(),
                 &mut np_dims as *mut npyffi::PyArray_Dims,
                 check_ref,
@@ -886,15 +1137,15 @@ impl<T: Element> PyArray<T, Ix2> {
             return Err(FromVecError::new(v.len(), last_len));
         }
         let dims = [v.len(), last_len];
-        let array = Self::new(py, dims, false);
         unsafe {
+            let array = Self::new(py, dims, false);
             for (y, vy) in v.iter().enumerate() {
                 for (x, vyx) in vy.iter().enumerate() {
-                    *array.uget_mut([y, x]) = vyx.clone();
+                    array.uget_raw([y, x]).write(vyx.clone());
                 }
             }
+            Ok(array)
         }
-        Ok(array)
     }
 }
 
@@ -928,17 +1179,17 @@ impl<T: Element> PyArray<T, Ix3> {
             return Err(FromVecError::new(v.len(), len3));
         }
         let dims = [v.len(), len2, len3];
-        let array = Self::new(py, dims, false);
         unsafe {
+            let array = Self::new(py, dims, false);
             for (z, vz) in v.iter().enumerate() {
                 for (y, vzy) in vz.iter().enumerate() {
                     for (x, vzyx) in vzy.iter().enumerate() {
-                        *array.uget_mut([z, y, x]) = vzyx.clone();
+                        array.uget_raw([z, y, x]).write(vzyx.clone());
                     }
                 }
             }
+            Ok(array)
         }
-        Ok(array)
     }
 }
 
@@ -949,7 +1200,7 @@ impl<T: Element, D> PyArray<T, D> {
     /// use numpy::PyArray;
     /// pyo3::Python::with_gil(|py| {
     ///     let pyarray_f = PyArray::arange(py, 2.0, 5.0, 1.0);
-    ///     let pyarray_i = PyArray::<i64, _>::new(py, [3], false);
+    ///     let pyarray_i = unsafe { PyArray::<i64, _>::new(py, [3], false) };
     ///     assert!(pyarray_f.copy_to(pyarray_i).is_ok());
     ///     assert_eq!(pyarray_i.readonly().as_slice().unwrap(), &[2, 3, 4]);
     /// });
@@ -957,7 +1208,7 @@ impl<T: Element, D> PyArray<T, D> {
     pub fn copy_to<U: Element>(&self, other: &PyArray<U, D>) -> PyResult<()> {
         let self_ptr = self.as_array_ptr();
         let other_ptr = other.as_array_ptr();
-        let result = unsafe { PY_ARRAY_API.PyArray_CopyInto(other_ptr, self_ptr) };
+        let result = unsafe { PY_ARRAY_API.PyArray_CopyInto(self.py(), other_ptr, self_ptr) };
         if result == -1 {
             Err(PyErr::fetch(self.py()))
         } else {
@@ -978,10 +1229,10 @@ impl<T: Element, D> PyArray<T, D> {
     /// ```
     pub fn cast<'py, U: Element>(&'py self, is_fortran: bool) -> PyResult<&'py PyArray<U, D>> {
         let ptr = unsafe {
-            let dtype = U::get_dtype(self.py());
             PY_ARRAY_API.PyArray_CastToType(
+                self.py(),
                 self.as_array_ptr(),
-                dtype.into_ptr() as _,
+                U::get_dtype(self.py()).into_dtype_ptr(),
                 if is_fortran { -1 } else { 0 },
             )
         };
@@ -1033,6 +1284,7 @@ impl<T: Element, D> PyArray<T, D> {
         let mut np_dims = dims.to_npy_dims();
         let ptr = unsafe {
             PY_ARRAY_API.PyArray_Newshape(
+                self.py(),
                 self.as_array_ptr(),
                 &mut np_dims as *mut npyffi::PyArray_Dims,
                 order,
@@ -1064,10 +1316,11 @@ impl<T: Element + AsPrimitive<f64>> PyArray<T, Ix1> {
     pub fn arange(py: Python, start: T, stop: T, step: T) -> &Self {
         unsafe {
             let ptr = PY_ARRAY_API.PyArray_Arange(
+                py,
                 start.as_(),
                 stop.as_(),
                 step.as_(),
-                T::npy_type() as i32,
+                T::get_dtype(py).num(),
             );
             Self::from_owned_ptr(py, ptr)
         }
@@ -1076,7 +1329,9 @@ impl<T: Element + AsPrimitive<f64>> PyArray<T, Ix1> {
 
 #[cfg(test)]
 mod tests {
-    use super::PyArray;
+    use super::*;
+
+    use std::ops::Range;
 
     #[test]
     fn test_get_unchecked() {
@@ -1095,4 +1350,48 @@ mod tests {
             array.to_dyn().to_owned_array();
         })
     }
+
+    #[test]
+    fn test_hasobject_flag() {
+        use super::ToPyArray;
+        use pyo3::{py_run, types::PyList, Py, PyAny};
+
+        pyo3::Python::with_gil(|py| {
+            let a = ndarray::Array2::from_shape_fn((2, 3), |(_i, _j)| PyList::empty(py).into());
+            let arr: &PyArray<Py<PyAny>, _> = a.to_pyarray(py);
+            py_run!(py, arr, "assert arr.dtype.hasobject");
+        });
+    }
+
+    struct InsincereIterator(Range<usize>, usize);
+
+    impl Iterator for InsincereIterator {
+        type Item = usize;
+
+        fn next(&mut self) -> Option<Self::Item> {
+            self.0.next()
+        }
+    }
+
+    impl ExactSizeIterator for InsincereIterator {
+        fn len(&self) -> usize {
+            self.1
+        }
+    }
+
+    #[test]
+    #[should_panic]
+    fn from_exact_iter_too_short() {
+        Python::with_gil(|py| {
+            PyArray::from_exact_iter(py, InsincereIterator(0..3, 5));
+        });
+    }
+
+    #[test]
+    #[should_panic]
+    fn from_exact_iter_too_long() {
+        Python::with_gil(|py| {
+            PyArray::from_exact_iter(py, InsincereIterator(0..5, 3));
+        });
+    }
 }
